﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using System.Threading;
using System.Diagnostics;

namespace AsyncRecursiveMethods {
	class Program {
		static int CurrentThreadId {
			get { return Thread.CurrentThread.ManagedThreadId; }
		}

		static async Task<int> f(int n, Func<int, Task<int>> decTaskFactory, int indent) {
			Console.Write("[StackTrace.Length = {3}] {2} Thread {0}: f({1}) begin", CurrentThreadId, n, new string('-', indent), new StackTrace(false).FrameCount);
			
			int result = 1;
			if (n == 0) {
				Console.WriteLine();
			} else {
				Console.WriteLine(" will ask for n-1 [before 1st await]");

				int nMinusOne = await decTaskFactory(n);
				Console.WriteLine("[StackTrace.Length = {3}] {2} Thread {0}: f({1}) got n-1 [after 1st await], will ask for f(n-1) [before 2nd await]", CurrentThreadId, n, new string('-', indent), new StackTrace(false).FrameCount);

				int nextF = await f(nMinusOne, decTaskFactory, indent + 1);
				Console.WriteLine("[StackTrace.Length = {3}] {2} Thread {0}: f({1}) got f(n-1) [after 2nd await], will compute n * f(n-1) synchronously", CurrentThreadId, n, new string('-', indent), new StackTrace(false).FrameCount);

				result = n * nextF;
			}

			Console.WriteLine("[StackTrace.Length = {3}] {2} Thread {0}: f({1}) end", CurrentThreadId, n, new string('-', indent), new StackTrace(false).FrameCount);
			return result;
		}

		static void Main(string[] args) {
			const int TheN = 10; // 5000;

			Console.WriteLine("Thread {0}: Main before f(..., with async adder)", CurrentThreadId);

			Task<int> futureF = f(TheN, i => Task.Run(() => { Thread.Sleep(100); return i - 1; }), indent: 1);

			Console.WriteLine("Thread {0}: Main got future f(..., with async adder), will wait for it ...", CurrentThreadId);

			futureF.Wait();

			Console.WriteLine("Thread {0}: Main got future f(..., with async adder) result -> {1}", CurrentThreadId, futureF.Result);

			//
			//
			//

			Console.WriteLine();

			//
			//
			//

			Console.WriteLine("Thread {0}: Main before f(..., with sync completed adder)", CurrentThreadId);

			Task<int> futureF2 = f(TheN, i => Task.FromResult(i - 1), indent: 1);

			Console.WriteLine("Thread {0}: Main got future f(..., with sync completed adder), will wait for it ...", CurrentThreadId);

			futureF2.Wait();

			Console.WriteLine("Thread {0}: Main got future f(..., with sync completed adder) result -> {1}", CurrentThreadId, futureF2.Result);

			//
			//
			//

			Console.WriteLine();

			//
			//
			//

			{
				Task initalTask = new Task(() => {
					Console.WriteLine("[StackTrace.Length = {0}] Thread {1}: Initial task is running and going to sleep ...", new StackTrace(false).FrameCount, CurrentThreadId);
					Thread.Sleep(1000);
					Console.WriteLine("[StackTrace.Length = {0}] Thread {1}: Initial task is running again and going to complete ...", new StackTrace(false).FrameCount, CurrentThreadId);
				});

				var prevTask = initalTask;
				for (int i = 0; i < TheN; i++) {
					int continuationId = i;
					var newTask = prevTask.ContinueWith((t) => {
						// Thread.Sleep(100);
						Console.WriteLine("[StackTrace.Length = {0}] Thread {1} <<< Continuation #{2} is running and going to complete ...", new StackTrace(false).FrameCount, CurrentThreadId, continuationId);
					});

					for (int j = 0; j < 3; j++) {
						int subId = j;

						prevTask.ContinueWith((t) => {
							// Thread.Sleep(100);
							Console.WriteLine("[StackTrace.Length = {0}] Thread {1} - Continuation #{2} - SIDE {3} is running and going to complete ...", new StackTrace(false).FrameCount, CurrentThreadId, continuationId, subId);
						});
					}

					prevTask = newTask;
				}

				Console.WriteLine("Thread {0}: Main is starting Initial task ...", CurrentThreadId);
				initalTask.Start();

				Thread.Sleep(500);
				Console.WriteLine("Thread {0}: Main will wait for last continuation ...", CurrentThreadId);
				prevTask.Wait();

				Console.WriteLine("Thread {0}: Main done waiting for last continuation ...", CurrentThreadId);
			}

			//
			//
			//

			Console.WriteLine();

			//
			//
			//

			{
				Task initalTask = new Task(() => {
					Console.WriteLine("[StackTrace.Length = {0}] Thread {1}: Initial task is running and going to sleep ...", new StackTrace(false).FrameCount, CurrentThreadId);
					Thread.Sleep(1000);
					Console.WriteLine("[StackTrace.Length = {0}] Thread {1}: Initial task is running again and going to complete ...", new StackTrace(false).FrameCount, CurrentThreadId);
				});

				var prevTask = initalTask;
				for (int i = 0; i < TheN; i++) {
					int continuationId = i;
					var newTask = prevTask.ContinueWith((t) => {
						Console.WriteLine("[StackTrace.Length = {0}] Thread {1} <<< Continuation #{2} is running and going to complete ...", new StackTrace(false).FrameCount, CurrentThreadId, continuationId);
					});

					prevTask = newTask;
				}

				Console.WriteLine("Thread {0}: Main is starting Initial task ...", CurrentThreadId);
				initalTask.Start();

				Thread.Sleep(500);
				Console.WriteLine("Thread {0}: Main will wait for last continuation ...", CurrentThreadId);
				prevTask.Wait();

				Console.WriteLine("Thread {0}: Main done waiting for last continuation ...", CurrentThreadId);
			}

		}
	}
}
